# -*- coding: utf-8 -*-
# PyMoneyStats - On-Line Budget Manager
#
# Copyright (C) 2012 - 2015 Mark Hall
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
u"""

.. moduleauthor:: Mark Hall <mark.hall@mail.room3b.eu>
"""
import datetime
import formencode
import re
import transaction

from csv import DictReader
from formencode import validators
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
from pywebtools import text
from pywebtools.renderer import render
from sqlalchemy import and_, desc, asc

from pymoneystats import helpers
from pymoneystats.models import (DBSession, LineItem, Tag)

ISO_DATE = r'[0-9]{4}-[0-9]{2}-[0-9]{2}'


class CSVValidator(formencode.FancyValidator):

    messages = {'notacsv': 'The file you have selected is not a valid CSV file',
                'missingfields': '''The CSV file you selected does not contain all the required columns (date, title,
                amount, direction, category, and tags)''',
                'invaliddateformat': 'A date value is not in the requried format YYYY-MM-DD',
                'invaliddate': 'A date value is outside the valid range',
                'invalidamount': 'An amount value is not a valid number'}
    
    def _validate_python(self, value, state):
        try:
            reader = DictReader(value.file)
            fieldnames = reader.fieldnames
            for fieldname in ['date', 'title', 'amount', 'direction', 'category', 'tags']:
                if fieldname not in fieldnames:
                    raise formencode.Invalid(self.message('missingfields', state), value, state)
            for line in reader:
                try:
                    validators.Number(not_empty=True).to_python(line['amount'], state)
                except formencode.Invalid as e:
                    raise formencode.Invalid(self.message('invalidamount', value), value, state)
                match = re.match(ISO_DATE, line['date'])
                if not match:
                    raise formencode.Invalid(self.message('invaliddate', state), value, state)
                else:
                    try:
                        datetime.datetime.strptime(line['date'], '%Y-%m-%d')
                    except Exception as e:
                        print(e)
                        raise formencode.Invalid(self.message('invaliddate', state), value, state)
            value.file.seek(0)
        except Exception as e:
            print(e)
            if isinstance(e, formencode.Invalid):
                raise e
            else:
                raise formencode.Invalid(self.message('notacsv', value), value, state)


class ImportSchema(formencode.Schema):

    source = formencode.All(validators.FieldStorageUploadConverter(not_empty=True),
                            CSVValidator())


@view_config(route_name='importexport.index')
@render({'text/html': 'importexport/index.html'})
def index(request):
    return {}


@view_config(route_name='importexport.export')
@render({'text/csv': True})
def export(request):
    request.response.content_disposition = 'attachment;filename="pymoneystats.csv"'
    dbsession = DBSession()
    rows = []
    for item in dbsession.query(LineItem).order_by(LineItem.date, LineItem.id):
        rows.append({'date': item.date.isoformat(),
                     'title': item.title,
                     'amount': item.amount,
                     'direction': item.direction,
                     'category': item.category,
                     'tags': ','.join([t.tag for t in item.tags])})
    return {'columns': ['date', 'title', 'amount', 'direction', 'category', 'tags'], 'rows': rows}


@view_config(route_name='importexport.import')
@render({'text/html': 'importexport/index.html'})
def import_file(request):
    try:
        params = ImportSchema().to_python(request.POST)
        dbsession = DBSession()
        reader = DictReader(params['source'].file)
        with transaction.manager:
            for line in reader:
                line_item = LineItem(date=datetime.datetime.strptime(line['date'], '%Y-%m-%d'),
                                     title=line['title'],
                                     amount=float(line['amount']),
                                     direction=line['direction'],
                                     category=line['category'])
                tags = [t for t in line['tags'].split(',') if t.strip()]
                for tag_name in tags:
                    tag = dbsession.query(Tag).filter(Tag.tag==tag_name).first()
                    if not tag:
                        tag = Tag(tag=tag_name)
                        dbsession.add(tag)
                    line_item.tags.append(tag)
                dbsession.add(line_item)
        raise HTTPFound(request.route_url('items.index'))
    except formencode.Invalid as e:
        print(e)
        e.params = request.POST
        return {'e': e}
