# -*- coding: utf-8 -*-
# PyMoneyStats - On-Line Budget Manager
#
# Copyright (C) 2012 Mark Hall
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
u"""

.. moduleauthor:: Mark Hall <mark.hall@mail.room3b.eu>
"""
import datetime
import formencode
import transaction

from formencode import validators, foreach
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
from pywebtools.renderer import render
from sqlalchemy import desc, func, and_

from pymoneystats import helpers
from pymoneystats.models import (DBSession, LineItem, Tag)

class IndexPaginationSchema(formencode.Schema):
    
    year = validators.Number(if_missing=None, if_empty=None)
    month = validators.Number(if_missing=None, if_empty=None, min_value=1, max_value=12)
    filter_category = foreach.ForEach(validators.UnicodeString(if_empty=None), if_missing=None)
    filter_tag = foreach.ForEach(validators.UnicodeString(if_empty=None), if_missing=None)
    
class LineItemSchema(formencode.Schema):
    
    title = validators.UnicodeString(not_empty=True)
    amount = validators.Number(not_empty=True)
    category = validators.UnicodeString(not_empty=True)
    date = validators.DateConverter(month_style='dd/mm/yyyy', not_empty=True)
    direction = validators.UnicodeString(not_empty=True)
    tags = validators.UnicodeString(if_empty='')

class AutocompleteSchema(formencode.Schema):
    
    field = validators.UnicodeString(not_empty=True)
    term = validators.UnicodeString(not_empty=True)
    
class SuggestSchema(formencode.Schema):
    
    title = validators.UnicodeString(not_empty=True)
    
def get_request_parameter(request, name, default=None):
    if name in request.params:
        return request.params[name]
    else:
        return default
    
@view_config(route_name='items.index')
@render({'text/html': 'items/index.html'})
def index(request):
    dbsession = DBSession()
    items = dbsession.query(LineItem).order_by(desc(LineItem.date), desc(LineItem.id))
    last_entry = dbsession.query(LineItem.date).order_by(desc(LineItem.date)).first()
    params = IndexPaginationSchema().to_python(request.params)
    current_month = None
    previous_month = None
    next_month = None
    if params['year'] and params['month']:
        end_year = params['year']
        end_month = params['month'] + 1
        if end_month > 12:
            end_year = end_year + 1
            end_month = 1
        current_month = datetime.date(params['year'], params['month'], 1)
        next_month = datetime.date(end_year, end_month, 1)
        items = items.filter(and_(LineItem.date >= current_month,
                                  LineItem.date < next_month))
    elif last_entry:
        current_month = datetime.date(last_entry.date.year, last_entry.date.month, 1)
        if current_month.month == 12:
            next_month = datetime.date(current_month.year + 1, 1, 1)
        else:
            next_month = datetime.date(current_month.year, current_month.month + 1, 1)
        items = items.filter(LineItem.date >= current_month)
    if current_month:
        if current_month.month == 1:
            previous_month = datetime.date(current_month.year - 1, 12, 1)
        else:
            previous_month = datetime.date(current_month.year, current_month.month - 1, 1)
    categories = set()
    tags = set()
    for item in items:
        categories.add(item.category)
        for tag in item.tags:
            tags.add(tag.tag)
    categories = list(categories)
    categories.sort()
    tags = list(tags)
    tags.sort()
    selected_categories = []
    if params['filter_category']:
        selected_categories = [f for f in params['filter_category'] if f in categories]
        if selected_categories:
            items = items.filter(LineItem.category.in_(selected_categories))
    selected_tags = []
    if params['filter_tag']:
        selected_tags = [t for t in params['filter_tag'] if t in tags]
        if selected_tags:
            items = items.join(LineItem.tags).filter(Tag.tag.in_(selected_tags))
    return {'items': items,
            'last_entry': last_entry,
            'current_month': current_month,
            'previous_month': previous_month,
            'next_month': next_month,
            'categories': categories,
            'selected_categories': selected_categories,
            'tags': tags,
            'selected_tags': selected_tags}

@view_config(route_name='items.add')
@render({'text/html': 'items/add.html'})
def add(request):
    dbsession = DBSession()
    last_entry = dbsession.query(LineItem.date).order_by(desc(LineItem.date)).first()
    if request.method == 'POST':
        try:
            params = LineItemSchema().to_python(request.POST)
        except formencode.Invalid as e:
            e.params = request.POST
            return {'e': e,
                    'last_entry': last_entry}
        with transaction.manager:
            item = LineItem(title=params['title'],
                            amount=params['amount'],
                            category=params['category'],
                            date=params['date'],
                            direction=params['direction'])
            tags = params['tags'].split(',')
            for tag_value in tags:
                tag_value = tag_value.strip()
                if tag_value != '':
                    tag = dbsession.query(Tag).filter(Tag.tag==tag_value).first()
                    if tag:
                        item.tags.append(tag)
                    else:
                        item.tags.append(Tag(tag=tag_value))
            dbsession.add(item)
        last_entry = dbsession.query(LineItem.date).order_by(desc(LineItem.date)).first()
    return {'last_entry': last_entry}

@view_config(route_name='items.edit')
@render({'text/html': 'items/edit.html'})
def edit(request):
    dbsession = DBSession()
    item = dbsession.query(LineItem).filter(LineItem.id==request.matchdict['iid']).first()
    if request.method == 'POST':
        try:
            params = LineItemSchema().to_python(request.POST)
        except formencode.Invalid as e:
            e.params = request.POST
            return {'e': e,
                    'item': item}
        with transaction.manager:
            item = dbsession.query(LineItem).filter(LineItem.id==request.matchdict['iid']).first()
            item.title = params['title']
            item.amount = params['amount']
            item.category = params['category']
            item.date = params['date']
            item.direction = params['direction']
            item.tags = []
            tags = params['tags'].split(',')
            for tag_value in tags:
                tag_value = tag_value.strip()
                if tag_value != '':
                    tag = dbsession.query(Tag).filter(Tag.tag==tag_value).first()
                    if tag:
                        item.tags.append(tag)
                    else:
                        item.tags.append(Tag(tag=tag_value))
            dbsession.add(item)
        raise HTTPFound(request.route_url('items.index'))
    return {'item': item}

@view_config(route_name='items.delete')
@render({'text/html': 'items/delete.html'})
def delete(request):
    dbsession = DBSession()
    item = dbsession.query(LineItem).filter(LineItem.id==request.matchdict['iid']).first()
    if request.method == 'POST':
        with transaction.manager:
            dbsession.delete(item)
        raise HTTPFound(request.route_url('items.index'))
    return {'item': item}


@view_config(route_name='items.autocomplete')
@render({'application/json': True})
def autocomplete(request):
    try:
        params = AutocompleteSchema().to_python(request.params)
        dbsession = DBSession()
        if params['field'] == 'title':
            query = dbsession.query(LineItem.title, func.count(LineItem.title).label('c')).filter(LineItem.title.ilike('%s%%' % (params['term']))).group_by(LineItem.title)
        elif params['field'] == 'category':
            query = dbsession.query(LineItem.category, func.count(LineItem.category).label('c')).filter(LineItem.category.ilike('%s%%' % (params['term']))).group_by(LineItem.category)
        elif params['field'] == 'tag':
            query = dbsession.query(Tag.tag, func.count(Tag.tag).label('c')).join(Tag.line_items).filter(Tag.tag.ilike('%s%%' % (params['term']))).group_by(Tag.tag)
        query = query.order_by(desc('c')).limit(5)
        return {'items': [{'title': helpers.text.title(v), 'value': v} for (v, _) in query]}
    except formencode.Invalid:
        return {'items': []}

@view_config(route_name='items.suggest')
@render({'application/json': True})
def suggest(request):
    try:
        params = SuggestSchema().to_python(request.params)
        dbsession = DBSession()
        values = {'category': {},
                  'amount': {},
                  'tags': {},
                  'direction': {}}
        total = 0.0
        last_entry_date = dbsession.query(LineItem.date).order_by(desc(LineItem.date)).first()
        if last_entry_date:
            last_entry_date = last_entry_date.date
        else:
            last_entry_date = datetime.datetime.now()
        for line_item in dbsession.query(LineItem).filter(LineItem.title==params['title']).filter(LineItem.date>=last_entry_date - datetime.timedelta(days=180)):
            if line_item.category in values['category']:
                values['category'][line_item.category] = values['category'][line_item.category] + 1
            else:
                values['category'][line_item.category] = 1
            if line_item.amount in values['amount']:
                values['amount'][line_item.amount] = values['amount'][line_item.amount] + 1
            else:
                values['amount'][line_item.amount] = 1
            if line_item.direction in values['direction']:
                values['direction'][line_item.direction] = values['direction'][line_item.direction] + 1
            else:
                values['direction'][line_item.direction] = 1
            for tag in line_item.tags:
                if tag.tag in values['tags']:
                    values['tags'][tag.tag] = values['tags'][tag.tag] + 1
                else:
                    values['tags'][tag.tag] = 1
            total = total + 1
        suggestions = {}
        for field in ['category', 'amount', 'direction', 'tags']:
            for value, count in values[field].items():
                if count / total > 0.66:
                    if field in suggestions:
                        if isinstance(suggestions[field], list):
                            suggestions[field].append(value)
                        else:
                            suggestions[field] = [suggestions[field], value]
                    else:
                        suggestions[field] = value
        return suggestions
    except formencode.Invalid:
        return {}
